"""
==========
Engine
==========

Engine runs the simulation.
"""

import cProfile
import pstats
import os
import logging as log
import pprint
import re
from typing import (
    Any, Dict, Optional, Union, Tuple, Callable, Iterable, List,
    cast, Sequence)
import math
import datetime
import time as clock
import uuid
import warnings

import networkx as nx
import numpy as np
import pytest

from vivarium.core.store import (
    hierarchy_depth, Store, generate_state, view_values)
from vivarium.core.emitter import get_emitter, Emitter
from vivarium.core.process import (
    Process,
    ParallelProcess,
    Step,
)
from vivarium.core.composer import Composite
from vivarium.library.topology import (
    get_in,
    delete_in,
    assoc_path,
    inverse_topology,
    normalize_path,
)
from vivarium.library.dict_utils import apply_func_to_leaves
from vivarium.core.types import (
    HierarchyPath, Topology, State, Update, Processes, Steps,
    Flow, Schema)

pretty = pprint.PrettyPrinter(indent=2)


def pp(x: Any) -> None:
    """Print ``x`` in a pretty format."""
    pretty.pprint(x)


def pf(x: Any) -> str:
    """Format ``x`` for display."""
    return pretty.pformat(x)


log.basicConfig(level=os.environ.get("LOGLEVEL", log.WARNING))


def starts_with(
        a_list: HierarchyPath,
        sub: HierarchyPath,
) -> bool:
    """Check whether one path is a prefix of another.

    Args:
        a_list: Path to look for prefix in.
        sub: Prefix.

    Returns:
        True if ``sub`` is a prefix of ``a_list``; False otherwise.
    """
    return len(sub) <= len(a_list) and all(
        a_list[i] == el
        for i, el in enumerate(sub))


def invert_topology(
        update: Update,
        args: Tuple[HierarchyPath, Topology],
) -> State:
    """Wrapper function around ``inverse_topology``.

    Wraps :py:func:`vivarium.library.topology.inverse_topology`.

    Updates are produced relative to the process that produced them. To
    transform them such that they are relative to the root of the
    simulation hierarchy, this function "inverts" a topology.

    Args:
        update: The update.
        args: Tuple of the path to which the update is relative and the
            topology.

    Returns:
        The update, relative to the root of ``path``.
    """
    path, topology = args
    return inverse_topology(path[:-1], update, topology)


def timestamp(dt: Optional[Any] = None) -> str:
    """Get a timestamp of the form ``YYYYMMDD.HHMMSS``.

    Args:
        dt: Datetime object to generate timestamp from. If not
            specified, the current time will be used.

    Returns:
        Timestamp.
    """
    if not dt:
        dt = datetime.datetime.now()
    return "%04d%02d%02d.%02d%02d%02d" % (
        dt.year, dt.month, dt.day,
        dt.hour, dt.minute, dt.second)


def empty_front(t: float) -> Dict[str, Union[float, dict]]:
    return {
        'time': t,
        'update': {}}


class Defer:
    def __init__(
            self,
            defer: Any,
            f: Callable,
            args: Tuple,
    ) -> None:
        """Allows for delayed application of a function to an update.

        The object simply holds the provided arguments until it's time
        for the computation to be performed. Then, the function is
        called.

        Args:
            defer: An object with a ``.get_command_result()`` method
                whose output will be passed to the function. For
                example, the object could be an
                :py:class:`vivarium.core.process.Process` object whose
                ``.get_command_result()`` method will return the process
                update.
            function: The function. For example,
                :py:func:`invert_topology` to transform the returned
                update.
            args: Passed as the second argument to the function.
        """
        self.defer = defer
        self.f = f
        self.args = args

    def get(self) -> Update:
        """Perform the deferred computation.

        Returns:
            The result of calling the function.
        """
        return self.f(
            self.defer.get_command_result(),
            self.args)


class EmptyDefer(Defer):
    def __init__(self) -> None:
        def function(update: dict, _: tuple) -> dict:
            return update
        args = ()
        super().__init__(None, function, args)

    def get(self) -> Update:
        return {}


class _StepGraph:
    """A dependency graph of :term:`steps`.

    A step is just a Process object that has dependencies on other
    steps. Unlike processes, which can be run in any order, steps run
    every timestep. In a given timestep, each step must not run until
    all its dependency steps have run and had their updates applied.

    Note that the constructor uses any provided arguments without
    copying them.

    Attributes:
        graph: A NetworkX DiGraph with an edge for each dependency
            relationship and a node for each step path. If the step at
            path ``a`` depends on the step at path ``b``, then the graph
            will contain an edge from ``b`` to ``a``. This means that a
            topological sort of the graph produces a valid runtime order
            for the step. Note that the graph must be a DAG.
        sequential_steps: A list of paths for steps that should run
            sequentially and before the steps in ``graph``. This is
            where we store legacy :term:`derivers` for
            backwards-compatibility.
    """

    def __init__(
            self,
            graph: Optional[nx.DiGraph] = None,
            sequential_steps: Optional[List[HierarchyPath]] = None
            ) -> None:
        self._graph = graph or nx.DiGraph()
        self._sequential_steps: List[HierarchyPath] = (
            sequential_steps or [])

    def _validate(self) -> None:
        if not nx.is_directed_acyclic_graph(self._graph):
            raise ValueError('Step graph must be a DAG.')
        graph_steps = set(self._graph.nodes)
        sequential_steps = set(self._sequential_steps)
        intersection = graph_steps & sequential_steps
        if intersection:
            raise ValueError(
                'self._graph and self._sequential_steps have '
                f'overlapping steps: {intersection}')

    def add(
            self,
            path: HierarchyPath,
            dependencies: Iterable[HierarchyPath]) -> None:
        """Add a step to the graph.

        Args:
            path: The step object's path in the hierarchy.
            dependencies: The hierarchy paths to each dependency.

        Raises:
            ValueError: If the graph produced by adding the step is not
                a DAG.
        """
        self._graph.add_node(path)
        for dependency in dependencies:
            self._graph.add_edge(dependency, path)
        self._validate()

    def add_sequential(
            self,
            path: HierarchyPath) -> None:
        """Add a step that is meant to run sequentially.

        Legacy steps (:term:`derivers`) were meant to run sequentially
        instead of being provided as a dependency graph. To support
        these legacy steps, this method adds a step that will run after
        all previously-added sequential steps and before any
        non-sequential steps.

        Args:
            path: The path to the step in the hierarchy.
        """
        self._sequential_steps.append(path)
        self._validate()

    def get_execution_layers(self) -> List[List[HierarchyPath]]:
        """Get step execution layers, with steps represnted by paths.

        An execution layer is a list of steps that can be executed in
        parallel. The graph's execution layers are an ordered list of
        these layers such that:

        * For a given layer, every step in the layer may be executed as
          soon as all steps in preceding layers have been executed.
        * Every step is in as early a layer as possible.

        In other words, the execution layers are the topological
        generations of the graph, prepended by any sequential steps,
        each in its own layer and in the order in which they were added.

        Returns:
            An ordered list of the execution layers, with each step
            represented by its path.
        """
        layers = nx.topological_generations(self._graph)
        to_return = [[step] for step in self._sequential_steps]
        to_return += [sorted(layer) for layer in layers]
        return to_return

    def remove(self, path: HierarchyPath) -> None:
        """Delete a step based on its path.

        Args:
            path: Hierarhcy path of the step to delete.
        """
        if path in self._sequential_steps:
            self._sequential_steps.remove(path)
            return
        to_delete = nx.algorithms.dag.descendants(self._graph, path)
        to_delete.add(path)
        for path_to_delete in to_delete:
            self._graph.remove_node(path_to_delete)

    def copy(self) -> '_StepGraph':
        """Create a copy of self.

        Returns:
            A new _StepGraph with a copy of self's graph.
        """
        new = self.__class__(
            self._graph.copy(), self._sequential_steps.copy())
        return new


class Engine:
    def __init__(
            self,
            composite: Optional[Composite] = None,
            processes: Optional[Processes] = None,
            steps: Optional[Steps] = None,
            flow: Optional[Flow] = None,
            topology: Optional[Topology] = None,
            store: Optional[Store] = None,
            initial_state: Optional[State] = None,
            experiment_id: Optional[str] = None,
            experiment_name: Optional[str] = None,
            metadata: Optional[dict] = None,
            description: str = '',
            emitter: Union[str, dict] = 'timeseries',
            store_schema: Optional[dict] = None,
            emit_topology: bool = True,
            emit_processes: bool = False,
            emit_config: bool = False,
            emit_step: float = 1,
            display_info: bool = True,
            progress_bar: bool = False,
            global_time_precision: Optional[int] = None,
            profile: bool = False,
            initial_global_time: float = 0,
    ) -> None:
        """Defines simulations

        Args:
            composite: A :term:`Composite`, which specifies the processes,
                steps, flow, and topology. This is an alternative to passing
                in processes and topology dict, which can not be loaded
                at the same time.
            processes: A dictionary that maps :term:`process` names to
                process objects. You will usually get this from the
                ``processes`` key of the dictionary from
                :py:meth:`vivarium.core.composer.Composer.generate`.
            steps: A dictionary that maps :term:`step` names to step
                objects. You will usually get this from the ``steps``
                key of the dictionary from
                :py:meth:`vivarium.core.composer.Composer.generate`.
            flow: A dictionary that maps :term:`step` names to sequences
                of paths to the steps that the step depends on. You will
                usually get this from the ``flow`` key of the dictionary
                from
                :py:meth:`vivarium.core.composer.Composer.generate`.
            topology: A dictionary that maps process names to
                sub-dictionaries. These sub-dictionaries map the
                process's port names to tuples that specify a path
                through the :term:`tree` from the :term:`compartment`
                root to the :term:`store` that will be passed to the
                process for that port.
            store: A pre-loaded Store. This is an alternative to passing
                in processes and topology dict, which can not be loaded
                at the same time. Note that if you provide this
                argument, you must ensure that all parallel processes
                (i.e. :py:class:`vivarium.core.process.Process` objects
                with the ``parallel`` attribute set to ``True``) are
                instances of
                :py:class:`vivarium.core.process.ParallelProcess`. This
                constructor converts parallel processes to
                ``ParallelProcess`` objects automatically if you do not
                provide this ``store`` argument.
            initial_state: By default an empty dictionary, this is the
                initial state of the simulation.
            experiment_id: A unique identifier for the experiment. A
                UUID will be generated if none is provided.
            metadata: A dictionary with additional data about the experiment,
                which is saved by the emitter with the configuration.
            description: A description of the experiment. A blank string
                by default.
            emitter: An emitter configuration which must conform to the
                specification in the documentation for
                :py:func:`vivarium.core.emitter.get_emitter`. The
                experiment ID will be added to the dictionary you
                provide as the value for the key ``experiment_id``.
            display_info: prints experiment info
            progress_bar: shows a progress bar
            global_time_precision: an optional int that sets the decimal
                precision of global_time. This is useful for remove floating-
                point rounding errors for the time keys of saved states.
            store_schema: An optional dictionary to expand the store hierarchy
                configuration, and also to turn emits on or off. The dictionary
                needs to be structured as a hierarchy, which will expand the
                existing store hierarchy. Setting an emit value for a branch
                node will set the emits of all the leaves to that value.
            emit_topology: If True, this will emit the topology with the
                configuration data.
            emit_processes: If True, this will emit the serialized
                processes with the configuration data.
            emit_config: If True, this will emit the serialized initial
                state with the configuration data.
            profile: Whether to profile the simulation with cProfile.
            initial_global_time: The initial time for the simulation.
                Useful when this Engine is part of a larger, older
                simulation.
        """
        self.profiler: Optional[cProfile.Profile] = None
        if profile:
            self.profiler = cProfile.Profile()
            self.profiler.enable()
        self.stats_objs: List[pstats.Stats] = []
        self.stats: Optional[pstats.Stats] = None

        self.experiment_id = experiment_id or str(uuid.uuid1())
        self.initial_state = initial_state or {}
        self.emit_step = emit_step

        # make the processes, topology, steps, flow, and store
        self._make_store(store, composite, processes, steps, flow, topology)

        # display settings
        self.experiment_name = experiment_name or self.experiment_id
        self.metadata = metadata
        self.description = description
        self.display_info = display_info
        self.global_time_precision = global_time_precision
        self.progress_bar = progress_bar
        self.time_created = timestamp()
        if self.display_info:
            self._print_display()

        # get a mapping of all paths to processes
        self.process_paths: Dict[HierarchyPath, Process] = {}
        self._step_graph = _StepGraph()
        self._step_paths: Dict[HierarchyPath, Process] = {}
        self._find_process_paths(self.processes, self.flow)
        self._find_step_paths(self.steps, self.flow)
        self._validate_steps_and_flow(self._step_paths, self.flow)

        # emitter settings
        emitter_config = emitter
        if isinstance(emitter_config, str):
            emitter_config = {'type': emitter_config}
        else:
            emitter_config = dict(emitter_config)
        emitter_config['experiment_id'] = self.experiment_id
        self.emitter: Emitter = get_emitter(emitter_config)

        # override emit settings in store
        if store_schema:
            self.state._apply_config(store_schema)

        # settings for self._emit_configuration()
        self.emit_topology = emit_topology
        self.emit_processes = emit_processes
        self.emit_config = emit_config

        # initialize global time
        self.global_time = initial_global_time

        # front tracks how far each process has been simulated in time,
        # and also holds the processes' updates before they are applied.
        self.front: Dict = {
            path: empty_front(self.global_time)
            for path in self.process_paths}

        # run the steps
        self.run_steps()

        # run the emitter
        self._emit_configuration()
        self._emit_store_data()

        # logging information
        log.info('experiment %s', str(self.experiment_id))

        log.info('\nPROCESSES:')
        log.info(pf(self.processes))

        log.info('\nTOPOLOGY:')
        log.info(pf(self.topology))

    @staticmethod
    def _validate_steps_and_flow(
            step_paths: Dict[HierarchyPath, Process],
            flow: Flow,
            path: HierarchyPath = tuple()) -> None:
        '''Check that every Step in flow is in steps.'''
        for key, sub_flow in flow.items():
            if isinstance(sub_flow, dict):
                Engine._validate_steps_and_flow(
                    step_paths, sub_flow, path + (key,))
            else:
                assert isinstance(sub_flow, list)
                for dependency in sub_flow:
                    dependency = path + dependency
                    if dependency not in step_paths:
                        raise ValueError(
                            f'Unknown dependency step {dependency} is '
                            'in the flow')

    def _make_store(
            self,
            store: Optional[Store] = None,
            composite: Optional[Composite] = None,
            processes: Optional[Processes] = None,
            steps: Optional[Steps] = None,
            flow: Optional[Flow] = None,
            topology: Optional[Topology] = None,
    ) -> None:
        """
        If a :term:`Store` is provided, retrieve the :term:`Processes`,
        :term:`Steps`, :term:`Flow`, and :term:`Topology`. If a
        :term:`Composite` or its attributes are provided, create a
        store. These are interchangeable.
        """
        if not store:
            if (processes and topology) or (steps and topology):
                self.processes = processes or {}
                self.steps = steps or {}
                self.flow = flow or {}
                self.topology = topology
            elif composite:
                self.processes = composite['processes']
                self.steps = composite['steps']
                self.flow = composite['flow']
                self.topology = composite['topology']
                self.initial_state = composite['state'] or self.initial_state
            else:
                raise ValueError(
                    'load either composite, store, or '
                    '(processes and topology) into Engine')

            self.processes = cast(
                Dict[str, Any],
                self._parallelize_processes(self.processes)
            )
            self.steps = cast(
                Dict[str, Any],
                self._parallelize_processes(self.steps)
            )

            # put the parallelized processes back in the composite.
            if composite:
                composite['processes'] = self.processes
                composite['steps'] = self.steps

            # initialize the store
            self.state: Store = generate_state(
                self.processes,
                self.topology,
                self.initial_state,
                self.steps,
                self.flow,
            )

        else:
            self.state = store
            self.state.set_value(self.initial_state)
            # build the processes' views
            self.state.build_topology_views()
            # get processes and topology from the store
            self.processes = self.state.get_processes()
            self.steps = self.state.get_steps() or {}
            self.flow = self.state.get_flow() or {}
            self.topology = self.state.get_topology()

    def _parallelize_processes(
            self, processes: Any) -> Union[dict, Process]:
        '''Replace parallel processes with ParallelProcess objects.'''
        if isinstance(processes, Process):
            if processes.parallel and not isinstance(
                    processes, ParallelProcess):
                processes = ParallelProcess(
                    processes, bool(self.profiler), self.stats_objs)
        elif isinstance(processes, dict):
            processes = {
                key: self._parallelize_processes(value)
                for key, value in processes.items()
            }
        else:
            raise AssertionError(f'Unrecognized collection: {processes}')
        return processes

    def _add_step_path(
            self,
            step: Step,
            path: HierarchyPath,
            # None if deriver, empty list if no dependencies.
            relative_dependencies: Optional[Sequence[HierarchyPath]],
    ) -> None:
        assert step.is_step()
        self._step_paths[path] = step
        if relative_dependencies is None:
            self._step_graph.add_sequential(path)
            return
        dependencies = [
            path + ('..',) + dep for dep in relative_dependencies]
        norm_dependencies = [
            normalize_path(dep) for dep in dependencies]
        self._step_graph.add(path, norm_dependencies)

    def _add_process_path(
            self,
            process: Process,
            path: HierarchyPath,
            flow: Flow,
    ) -> None:
        if process.is_step():
            # warnings.warn(
            #     f'Found a step {path} in the processes dict. This is '
            #     'deprecated. Steps should be specified in the steps '
            #     'dict instead.',
            #     category=FutureWarning,
            # )
            step = cast(Step, process)
            self._add_step_path(step, path, get_in(flow, path))
        else:
            self.process_paths[path] = process

    def _find_process_paths(
            self,
            processes: Processes,
            flow: Flow,
    ) -> None:
        tree = hierarchy_depth(processes)
        for path, process in tree.items():
            self._add_process_path(process, path, flow)

    def _find_step_paths(
            self,
            steps: Steps,
            flow: Flow,
    ) -> None:
        tree = hierarchy_depth(steps)
        for path, step in tree.items():
            self._add_step_path(step, path, get_in(flow, path))

    def _emit_configuration(self) -> None:
        """Emit experiment configuration."""
        data: Dict[str, Any] = {
            'time_created': self.time_created,
            'experiment_id': self.experiment_id,
            'name': self.experiment_name,
            'description': self.description,
            'metadata': self.metadata,
            'topology': self.topology
            if self.emit_topology else None,
            'processes': self.processes
            if self.emit_processes else None,
            'state': self.state.get_config()
            if self.emit_config else None,
        }
        emit_config: Dict[str, Any] = {
            'table': 'configuration',
            'data': data
        }
        self.emitter.emit(emit_config)

    def _emit_store_data(self) -> None:
        """Emit the current simulation state.
        Only variables with ``_emit=True`` are emitted.
        """
        data = self.state.emit_data()
        data.update({
            'time': self.global_time})
        emit_config = {
            'table': 'history',
            'data': data}
        self.emitter.emit(emit_config)

    def _process_state(
            self,
            path: HierarchyPath,
    ) -> Tuple[Store, State]:
        """Get the simulation state for a process's ``next_update``.

        Before computing an update, we have to collect the simulation
        variables the processes expects.

        Args:
            path: Path to the process.

        Returns:
            Tuple of the store at ``path`` and a collection of state
            variables in the form the process expects.
        """
        store = self.state.get_path(path)
        assert isinstance(store.value, Process)

        # translate the values from the tree structure into the form
        # that this process expects, based on its declared topology
        topology_view = store.topology_view
        assert topology_view is not None, \
            f"store at path {path} does not have a topology_view"
        states = view_values(topology_view)

        return store, states

    def _calculate_update(
            self,
            path: HierarchyPath,
            process: Process,
            interval: float
    ) -> Tuple[Defer, Store]:
        """Calculate a process's update.

        Args:
            path: Path to process.
            process: The process.
            interval: Timestep to compute update for.

        Returns:
            Tuple of the deferred update (relative to the root of
            ``path``) and the store at ``path``.
        """
        store, states = self._process_state(path)
        if process.update_condition(interval, states):
            return _process_update(
                path, process, store, states, interval)
        return (EmptyDefer(), store)

    def apply_update(
            self,
            update: Update,
            state: Store
    ) -> bool:
        """Apply an update to the simulation state.

        Args:
            update: The update to apply. Must be relative to ``state``.
            state: The store to which the update is relative (usually
                root of simulation state. We need this so to preserve
                the "perspective" from which the update was generated.

        Return:
            a bool indicating whether the topology_views expired.
        """

        if not update:
            return False

        (
            topology_updates, process_updates, step_updates,
            flow_updates, deletions, view_expire
        ) = self.state.apply_update(update, state)

        process_updates = [
            (path, self._parallelize_processes(process))
            for path, process in process_updates
        ]
        step_updates = [
            (path, self._parallelize_processes(step))
            for path, step in step_updates
        ]
        # Make sure the Store contains the parallelized processes.
        for path, process in process_updates:
            self.state.get_path(path).value = process
        for path, step in step_updates:
            self.state.get_path(path).value = step

        flow_update_dict = dict(flow_updates)

        if topology_updates:
            for path, topology_update in topology_updates:
                assoc_path(self.topology, path, topology_update)

        if flow_updates:
            for path, topology_update in flow_updates:
                assoc_path(self.flow, path, flow_updates)

        if process_updates:
            for path, process in process_updates:
                assoc_path(self.processes, path, process)
                self._add_process_path(process, path, {})

        if step_updates:
            for path, step in step_updates:
                dependencies = flow_update_dict.get(path)
                assoc_path(self.steps, path, step)
                self._add_step_path(step, path, dependencies)

        if deletions:
            for deletion in deletions:
                self._delete_path(deletion)

        return view_expire

    def _delete_path(
            self,
            deletion: HierarchyPath
    ) -> None:
        """Delete a store from the state.

        Args:
            deletion: Path to store to delete.
        """
        delete_in(self.processes, deletion)
        delete_in(self.steps, deletion)
        delete_in(self.topology, deletion)
        delete_in(self.flow, deletion)

        for path in list(self.process_paths.keys()):
            if starts_with(path, deletion):
                del self.process_paths[path]

        for path in list(self._step_paths):
            if starts_with(path, deletion):
                try:
                    self._step_graph.remove(path)
                except nx.exception.NetworkXError as e:
                    # The step might have been deleted already.
                    msg = f'The node {path} is not in the digraph.'
                    if e.args[0] != msg:
                        raise e
                del self._step_paths[path]

    def run_steps(self) -> None:
        """Run all the steps in the simulation."""
        layers = self._step_graph.get_execution_layers()
        for layer in layers:
            deferred_updates: List[Tuple[Defer, Store]] = []
            for path in layer:
                step = self._step_paths.get(path)
                if not step:
                    # Step was deleted by a previous step.
                    continue
                # Timestep shouldn't influence steps.
                # TODO(jerry): Do something cleaner than having
                #  generate_paths() add a schema attribute to the Deriver.
                #  PyCharm's type check reports:
                #    Type Process doesn't have expected attribute 'schema'
                update, store = self._calculate_update(
                    path, step, 0)
                deferred_updates.append((update, store))

            view_expire = False
            for update, store in deferred_updates:
                view_expire_update = self.apply_update(update.get(), store)
                view_expire = view_expire or view_expire_update

            if view_expire:
                self.state.build_topology_views()

    def _send_updates(
            self,
            update_tuples: list
    ) -> None:
        """Apply updates and run steps.

        Args:
            update_tuples: List of tuples ``(update, state)`` where
                ``state`` is the store from whose perspective the update
                was generated.
        """
        view_expire = False
        for update_tuple in update_tuples:
            update, state = update_tuple
            view_expire_update = self.apply_update(update.get(), state)
            view_expire = view_expire or view_expire_update

        if view_expire:
            self.state.build_topology_views()

        self.run_steps()

    def update(
            self,
            interval: float,
    ) -> None:
        """
        Run each process for the given interval and force them to complete
        at the end of the interval. See ``run_for`` for the keyword args.
        """
        clock_start = clock.time()
        self.run_for(
            interval=interval,
            force_complete=True,
        )
        self._check_complete()
        runtime = clock.time() - clock_start
        if self.display_info:
            _print_summary(runtime)

    def _check_complete(self) -> None:
        """Check that all processes completed"""
        for path, advance in self.front.items():
            assert advance['time'] == self.global_time, \
                f"the process at path {path} is at time {advance['time']} " \
                f"instead of completing at global time {self.global_time}"
            assert len(advance['update']) == 0, \
                f"the process at path {path} is an unapplied update"

    def _remove_deleted_processes(self) -> None:
        '''Remove deleted processes from the front.'''
        self.front = {
            path: progress
            for path, progress in self.front.items()
            if path in self.process_paths}

    def run_for(
            self,
            interval: float,
            force_complete: bool = False,
    ) -> None:
        """Run each process within the given interval and update their states.

        :py:meth:`run_for` gives the caller more control over the
        simulation loop than :py:meth:`update`. In particular, it may be
        called repeatedly within a caller-managed simulation loop
        without forcing processes to complete after each call (as would
        be the case with :py:meth:`update`). It is the responsibility
        of the caller to ensure that when :py:meth:`run_for` is called
        for the last time in the caller-managed simulation loop,
        `force_complete=True` so that all the processes finish at the
        end of the simulation.

        Args:
            interval: the amount of time to simulate the composite.
            force_complete: a bool indicating whether to force processes
                to complete at the end of the interval.
        """
        end_time = self.global_time + interval
        emit_time = self.global_time + self.emit_step

        while self.global_time < end_time or force_complete:
            full_step = math.inf
            self._remove_deleted_processes()

            # processes at quiet paths don't meet their execution condition,
            # but still advance in time
            quiet_paths = []

            # go through each process and find those that are able to update
            # based on their most recent update time being less than global time
            for path, process in self.process_paths.items():
                if path not in self.front:
                    self.front[path] = empty_front(self.global_time)
                process_time = self.front[path]['time']

                if process_time <= self.global_time:

                    # get the time step
                    store, states = self._process_state(path)
                    process_timestep = process.calculate_timestep(states)

                    if force_complete:
                        # force the process to complete at end_time
                        future = min(process_time + process_timestep, end_time)
                    else:
                        future = process_time + process_timestep
                    if self.global_time_precision is not None:
                        # set future time based on global_time_precision
                        future = round(future, self.global_time_precision)

                    if future <= end_time:

                        # calculate the update for this process
                        if process.update_condition(process_timestep, states):
                            update = _process_update(
                                path, process, store, states, process_timestep)

                            # update front, to be applied at its projected time
                            self.front[path]['time'] = future
                            self.front[path]['update'] = update

                            # absolute timestep
                            timestep = future - self.global_time
                            if timestep < full_step:
                                full_step = timestep
                        else:
                            # mark this path "quiet" so its time can be advanced
                            self.front[path]['update'] = (EmptyDefer(), store)
                            quiet_paths.append(path)
                    else:
                        # absolute timestep
                        timestep = future - self.global_time
                        if timestep < full_step:
                            full_step = timestep

                else:
                    # don't shoot past processes that didn't run this time
                    process_delay = process_time - self.global_time
                    if process_delay < full_step:
                        full_step = process_delay

            # apply updates based on process times in self.front
            if full_step == math.inf:
                # no processes ran, jump to next process
                next_event = end_time
                for path in self.front.keys():
                    if self.front[path]['time'] < next_event:
                        next_event = self.front[path]['time']
                self.global_time = next_event

            elif self.global_time + full_step <= end_time:
                # at least one process ran within the interval
                # increase the time, apply updates, and continue
                self.global_time += full_step

                # advance all quiet processes to current time
                for quiet in quiet_paths:
                    self.front[quiet]['time'] = self.global_time

                # apply updates that are behind global time
                updates = []
                paths = []
                for path, advance in self.front.items():
                    if advance['time'] <= self.global_time \
                            and advance['update']:
                        new_update = advance['update']
                        updates.append(new_update)
                        advance['update'] = {}
                        paths.append(path)

                self._send_updates(updates)

                # display and emit
                if self.progress_bar:
                    print_progress_bar(self.global_time, end_time)
                if self.emit_step == 1:
                    self._emit_store_data()
                elif emit_time <= self.global_time:
                    while emit_time <= self.global_time:
                        self._emit_store_data()
                        emit_time += self.emit_step

            else:
                # all processes have run past the interval
                self.global_time = end_time

            if force_complete and self.global_time == end_time:
                force_complete = False

    @staticmethod
    def _end_process_if_parallel(process: Process) -> None:
        if process.parallel:
            assert isinstance(process, ParallelProcess)
            process.end()

    def end(self) -> None:
        """Terminate all processes running in parallel.

        This MUST be called at the end of any simulation with parallel
        processes. This function also ends profiling and computes
        profiling stats, including stats from parallel sub-processes.
        These stats are stored in ``self.stats``.
        """
        apply_func_to_leaves(
            self.processes, self._end_process_if_parallel)
        apply_func_to_leaves(
            self.steps, self._end_process_if_parallel)
        if self.profiler:
            self.profiler.disable()
            total_stats = pstats.Stats(self.profiler)
            for stats in self.stats_objs:
                total_stats.add(stats)
            self.stats = total_stats

    def _print_display(self) -> None:
        """Print simulation metadata."""
        date, time = self.time_created.split('.')
        print('\nSimulation ID: {}'.format(self.experiment_id))
        print('Created: {} at {}'.format(
            date[4:6] + '/' + date[6:8] + '/' + date[0:4],
            time[0:2] + ':' + time[2:4] + ':' + time[4:6]))
        if self.experiment_name is not self.experiment_id:
            print('Name: {}'.format(self.experiment_name))
        if self.description:
            print('Description: {}'.format(self.description))


def print_progress_bar(
        iteration: float,
        total: float,
        decimals: float = 1,
        length: int = 50,
) -> None:
    """Create terminal progress bar

    Args:
        iteration: Current iteration
        total: Total iterations
        decimals: Positive number of decimals in percent complete
        length: Character length of bar
    """
    progress: str = ("{0:." + str(decimals) + "f}").format(total - iteration)
    filled_length: int = int(length * iteration // total)
    filled_bar = '█' * filled_length + '-' * (length - filled_length)
    print(
        f'\rProgress:|{filled_bar}| {progress}/{float(total)} '
        f'simulated seconds remaining    ', end='\r')
    # Print New Line on Complete
    if iteration == total:
        print()

def _print_summary(
        runtime: float
) -> None:
    """Print summary of simulation runtime."""
    if runtime < 1:
        print('Completed in {:.6f} seconds'.format(runtime))
    else:
        print('Completed in {:.2f} seconds'.format(runtime))

def _invoke_process(
        process: Process,
        interval: float,
        states: State,
) -> Any:
    """Trigger computation of a process's update.

    To allow processes to run in parallel, this function only
    triggers update computation. When the function exits,
    computation may not be complete.

    Args:
        process: The process.
        interval: The timestep for which to compute the update.
        states: The simulation state to pass to
            :py:meth:`vivarium.core.process.Process.next_update`.

    Returns:
        The deferred simulation update, for example a
        :py:class:`vivarium.core.process.ParallelProcess`.
    """
    process.send_command('next_update', (interval, states))
    return process

def _process_update(
        path: HierarchyPath,
        process: Process,
        store: Store,
        states: State,
        interval: float,
) -> Tuple[Defer, Store]:
    """Start generating a process's update.

    This function is similar to :py:func:`_invoke_process` except in
    addition to triggering the computation of the process's update
    (by calling ``_invoke_process``), it also generates a
    :py:class:`Defer` object to transform the update into absolute
    terms.

    Args:
        path: Path to process.
        process: The process.
        store: The store at ``path``.
        states: Simulation state to pass to process's
            ``next_update`` method.
        interval: Timestep for which to compute the update.

    Returns:
        Tuple of the deferred update (in absolute terms) and
        ``store``.
    """
    process = _invoke_process(
        process,
        interval,
        states)

    absolute = Defer(
        process,
        invert_topology,
        (path, store.topology))

    return absolute, store

def test_flow_with_extra_step() -> None:
    class StepA(Step):

        def ports_schema(self) -> Schema:
            return {
                'a': {'_default': 1}
            }

        def next_update(self, timestep: float, states: State) -> Update:
            return {}

    class ProcessB(Process):

        def ports_schema(self) -> Schema:
            return {
                'b': {'_default': 1}
            }

        def next_update(self, timestep: float, states: State) -> Update:
            return {}


    expected_error = re.escape(
        'Unknown dependency step (\'stepA2\',) is in the flow')
    with pytest.raises(Exception, match=expected_error):
        _ = Engine(
            processes={'procB': ProcessB()},
            steps={'stepA1': StepA()},
            topology={
                'stepA1': {'a': ('a',)},
                'procB': {'b': ('b',)},
            },
            flow={'stepA1': [('stepA2',)]},
        )

def test_flow_with_valid_steps() -> None:
    class StepA(Step):

        def ports_schema(self) -> Schema:
            return {
                'a': {'_default': 1}
            }

        def next_update(self, timestep: float, states: State) -> Update:
            return {}

    class ProcessB(Process):

        def ports_schema(self) -> Schema:
            return {
                'b': {'_default': 1}
            }

        def next_update(self, timestep: float, states: State) -> Update:
            return {}

    _ = Engine(
        processes={'procB': ProcessB()},
        steps={
            'stepA1': StepA(),
            'stepA2': StepA(),
        },
        topology={
            'stepA1': {'a': ('a',)},
            'stepA2': {'a': ('a',)},
            'procB': {'b': ('b',)},
        },
        flow={
            'stepA1': [('stepA2',)],
            'stepA2': [],
        },
    )

def test_numpy_schema_validation() -> None:
    class ProcessA(Process):

        defaults = {
            'ports_schema_dict': {
                'b': {'_default': 1}
            }
        }

        def __init__(self, parameters: Optional[dict] = None) -> None:
            super().__init__(parameters)
            self.ports_schema_dict = self.parameters['ports_schema_dict']

        def ports_schema(self) -> Schema:
            return self.ports_schema_dict

        def next_update(self, timestep: float, states: State) -> Update:
            return {}

    # Since procA1 and procA2 differ in their _divider schemas,
    # _check_schema_support_defaults should raise a warning
    with pytest.warns(UserWarning, match='Incompatible schema assignment'):
        _ = Engine(
            processes={
                'procA1': ProcessA(
                    {
                        'ports_schema_dict': {
                            'a': {
                                '_divider': {
                                    'divider': 'set_value',
                                    'config': {
                                        'value': np.ones(10)
                                    }
                                }
                            }
                        }
                    }
                ),
                'procA2': ProcessA(
                    {
                        'ports_schema_dict': {
                            'a': {
                                '_divider': {
                                    'divider': 'set_value',
                                    'config': {
                                        'value': np.zeros(10)
                                    }
                                }
                            }
                        }
                    }
                ),
            },
            topology={
                'procA1': {'a': ('a',)},
                'procA2': {'a': ('a',)},
            }
        )

    # Check that np.ones(10) is not considered equal to 1
    with pytest.raises(ValueError, match='The truth value of an array'):
        _ = Engine(
            processes={
                'procA1': ProcessA(
                    {
                        'ports_schema_dict': {
                            'a': {
                                '_divider': {
                                    'divider': 'set_value',
                                    'config': {
                                        'value': 1
                                    }
                                }
                            }
                        }
                    }
                ),
                'procA2': ProcessA(
                    {
                        'ports_schema_dict': {
                            'a': {
                                '_divider': {
                                    'divider': 'set_value',
                                    'config': {
                                        'value': np.ones(10)
                                    }
                                }
                            }
                        }
                    }
                ),
            },
            topology={
                'procA1': {'a': ('a',)},
                'procA2': {'a': ('a',)},
            }
        )

    # Since ProcA1 and ProcA2 differ in their _value schemas,
    # _check_schema should raise an exception
    with pytest.raises(ValueError, match='Incompatible schema assignment'):
        _ = Engine(
            processes={
                'procA1': ProcessA(
                    {
                        'ports_schema_dict': {
                            'a': {
                                '_value': np.ones(10)
                            }
                        }
                    }
                ),
                'procA2': ProcessA(
                    {
                        'ports_schema_dict': {
                            'a': {
                                '_value': np.zeros(10)
                            }
                        }
                    }
                ),
            },
            topology={
                'procA1': {'a': ('a',)},
                'procA2': {'a': ('a',)},
            }
        )

    # Matching schemas should raise no warnings or exceptions
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        _ = Engine(
                processes={
                    'procA1': ProcessA(
                        {
                            'ports_schema_dict': {
                                'a': {
                                    '_value': np.ones(10),
                                    '_divider': {
                                        'divider': 'set_value',
                                        'config': {
                                            'value': np.ones(10)
                                        }
                                    }
                                }
                            }
                        }
                    ),
                    'procA2': ProcessA(
                        {
                            'ports_schema_dict': {
                                'a': {
                                    '_value': np.ones(10),
                                    '_divider': {
                                        'divider': 'set_value',
                                        'config': {
                                            'value': np.ones(10)
                                        }
                                    }
                                }
                            }
                        }
                    ),
                },
                topology={
                    'procA1': {'a': ('a',)},
                    'procA2': {'a': ('a',)},
                }
            )



"""
===============
Burst Process
===============
"""

import os

from vivarium.core.engine import pp
from vivarium.core.process import (
    Step,
)
from vivarium.core.composer import Composer, Composite
from vivarium.core.engine import Engine
from vivarium.core.directories import PROCESS_OUT_DIR
from vivarium.composites.toys import ExchangeA
from vivarium.processes.timeline import TimelineProcess


NAME = 'burst'


class Burst(Step):
    """ Burst Process

    Remove a compartment when the state under the 'trigger' port is set to True.
    Move its inner stores to the outer.
    """
    name = NAME
    defaults = {
        'agent_id': 'DEFAULT'
    }

    def __init__(self, parameters=None):
        super().__init__(parameters)
        self.agent_id = self.parameters['agent_id']

    def ports_schema(self):
        return {
            'trigger': {
                '_default': False,
            },
            'inner': {
                '*': {}
            },
            'outer': {
                '*': {}
            },
            'compartment': {
                '*': {}
            },
        }

    def next_update(self, timestep, states):
        if states['trigger']:
            inner_states = states['inner']

            return {
                'inner': {
                    '_move': [{
                        # points to key in 'inner' port
                        'source': state,
                        # points to which port it will be moved
                        'target': 'outer',
                    } for state in inner_states],
                },
                'compartment': {
                    # remove self
                    '_delete': [self.agent_id]
                },
                'trigger': {
                    '_updater': 'set',
                    '_value': False},
            }

        else:
            return {}


# test
class ToyAgent(Composer):
    defaults = {
        'exchange': {
            'uptake_rate': 0.1},
        'inner_path': ('concentrations',),
        'outer_path': ('..', '..', 'concentrations'),
        'compartment_path': ('..', '..', 'agents'),
    }

    def generate_processes(self, config):
        agent_id = config['agent_id']
        return {
            'exchange': ExchangeA(config['exchange']),
            'burst': Burst({'agent_id': agent_id})}

    def generate_topology(self, config):
        return {
            'exchange': {
                'internal': config['inner_path'],
                'external': config['outer_path']},
            'burst': {
                'trigger': ('trigger',),
                'inner': config['inner_path'],
                'outer': config['outer_path'],
                'compartment': config['compartment_path'],
            }}


def test_burst(return_value=False):
    agent_1_id = '1'
    agent_2_id = '2'

    # initial state
    initial_a = 10
    initial_state = {
        'concentrations': {
            'A': initial_a
        },
        'agents': {
            agent_1_id: {
                'trigger': False,
                'concentrations': {
                    'A': 0
                },
                'agents': {
                    agent_2_id: {
                        'trigger': False,
                        'concentrations': {
                            'A': 0
                        },
                    }
                },
            }
        }
    }

    # timeline triggers burst for agent_s
    time_burst = 3
    time_total = 5
    timeline = [
        (0, {('agents', agent_1_id, 'agents', agent_2_id, 'trigger'): False}),
        (time_burst, {('agents', agent_1_id, 'agents', agent_2_id, 'trigger'): True}),
        (time_total, {})]

    # declare the hierarchy
    timeline = TimelineProcess({'timeline': timeline})
    agent1 = ToyAgent({'agent_id': agent_1_id})
    agent2 = ToyAgent({'agent_id': agent_2_id})

    composite = Composite()
    composite.merge(agent1.generate(path=('agents', agent_1_id)))
    composite.merge(agent2.generate(path=('agents', agent_1_id, 'agents', agent_2_id)))
    composite.merge(
        processes={'timeline': timeline},
        topology={'timeline': {
            'global': ('global',),
            'agents': ('agents',)}},
        state=initial_state
    )

    # configure experiment
    experiment = Engine(
        composite=composite,
    )

    pp(experiment.topology)
    pp(experiment.state.get_value())

    # run simulation
    experiment.update(time_total)
    output = experiment.emitter.get_data()
    experiment.end()  # end required for parallel processes

    # asserts total A is the same at the beginning and the end
    assert output[0.0]['concentrations']['A'] == initial_a
    assert output[5.0]['concentrations']['A'] + output[5.0]['agents'][agent_1_id]['concentrations']['A'] == initial_a

    if return_value:
        return output


def run_burst():
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    os.makedirs(out_dir, exist_ok=True)
    output = test_burst(return_value=True)
    pp(output)


if __name__ == '__main__':
    run_burst()



from typing import Any, Dict

from vivarium.core.process import Deriver


class DivideCondition(Deriver):
    """ Divide Condition Process """
    name = 'divide_condition'
    defaults: Dict[str, Any] = {}

    def __init__(self, parameters=None):
        super().__init__(parameters)
        self.threshold = self.parameters['threshold']

    def initial_state(self, config=None):
        return {}

    def ports_schema(self):
        return {
            'variable': {},
            'divide': {
                '_default': False,
                '_updater': 'set',
                '_divider': {
                    'divider': 'set_value',
                    'config': {
                        'value': False,
                    },
                },
            },
        }

    def next_update(self, timestep, states):
        if states['variable'] >= self.threshold:
            return {'divide': True}
        else:
            return {}



'''
===============================
Utilities for Formatting Output
===============================
'''

import inspect
import json
import numpy as np
import re

import pint

from vivarium.library.units import units
from vivarium.core.process import Process
from vivarium.core.composer import Composer


def _json_serialize(elem):
    if isinstance(elem, np.int64):
        return int(elem)
    if inspect.isfunction(elem):
        return '<function {}>'.format(elem.__name__)
    if inspect.isclass(elem):
        return str(elem)
    if isinstance(elem, (Composer, Process)):
        to_strip_regex = ' at 0x[0-9a-f]+>$'
        return re.sub(to_strip_regex, '>', repr(elem))
    if isinstance(elem, pint.Quantity):
        return str(elem)
    if isinstance(elem, pint.Unit):
        return repr(elem)
    return repr(elem)


def format_dict(d, sort_keys=True):
    '''Format a dict as a pretty string

    Aside from the normal JSON-serializable data types, data of type
    ``numpy.int64`` are supported.

    For example:

    >>> import numpy as np
    >>> d = {
    ...     'foo': {
    ...         'bar': 1,
    ...         '3.0': np.int64(5),
    ...     },
    ...     'a': 'hi!',
    ...     'quantity': 1 * units.fg,
    ...     'unit': units.fg,
    ... }
    >>> print(format_dict(d))
    {
        "a": "hi!",
        "foo": {
            "3.0": 5,
            "bar": 1
        },
        "quantity": "1 femtogram",
        "unit": "<Unit('femtogram')>"
    }

    Arguments:
        d: The dictionary to format
        sort_keys: Whether to sort the dictionary keys. This is useful
            for reproducible output.

    Returns:
        A string of the prettily-formatted dictionary

    '''
    return json.dumps(
        d, indent=4, default=_json_serialize, sort_keys=sort_keys)



"""
===========
Composition
===========

Helper functions for initializing and running experiments.
"""

import warnings
warnings.warn(
    "composition.py is no longer being maintained and will be "
    "removed in a future version of vivarium-core. Please use "
    "Engine directly to run simulations, and Composite.merge "
    "for composition", DeprecationWarning)

# typing
from typing import (
    Any, Dict, Optional)
from vivarium.core.types import (
    Topology, HierarchyPath)

from vivarium.core.process import Process
from vivarium.core.composer import Composer, Composite
from vivarium.core.engine import Engine
from vivarium.core.directories import (
    REFERENCE_DATA_DIR, BASE_OUT_DIR, TEST_OUT_DIR,
    PROCESS_OUT_DIR, COMPARTMENT_OUT_DIR, COMPOSITE_OUT_DIR,
    EXPERIMENT_OUT_DIR,
)
from vivarium.library.dict_utils import deep_merge

from vivarium.processes.timeline import TimelineProcess
from vivarium.processes.nonspatial_environment import \
    NonSpatialEnvironment

# toys for testing
from vivarium.composites.toys import (
    PoQo,
    ToyLinearGrowthDeathProcess,
    ExchangeA,
    ToyCompartment
)


# list of keys expected in experiment settings
experiment_config_keys = [
        'experiment_id',
        'experiment_name',
        'description',
        'initial_state',
        'emitter',
        'emit_step',
        'display_info',
        'progress_bar',
        'invoke',
    ]


# experiment loading functions
def add_timeline(
        processes: Dict[str, Process],
        topology: Dict[str, Topology],
        timeline: Dict[str, Any],
) -> None:
    """Add a timeline process to a composite

    Args:
        processes: with ``{'process_name': Process}``
        topology (dict): with ``{'process_name': topology mapping}``
        timeline (dict): with ``timeline`` key. An optional ``paths``
            key overrides the topology mapping from (port: path).
    """
    timeline_paths = timeline.get('paths', {})
    timeline_process = TimelineProcess(timeline)
    processes.update({
        TimelineProcess.name: timeline_process})

    # add topology
    timeline_topology = {
        port: timeline_paths.get(port, (port,))
        for port in timeline_process.ports()}
    topology.update({TimelineProcess.name: timeline_topology})


def add_environment(
        processes: Dict[str, Any],
        topology: Dict[str, Any],
        environment: Dict[str, Any],
) -> None:
    """Add a NonSpatialEnvironment to a composite

    Args:
        processes: with ``{'process_name': Process}``
        topology: with ``{'process_name': topology mapping}``
        environment: with ``environment`` key. An optional ``paths`` key
            overrides the topology mapping from (port: path).
    """

    overide_topology = environment.get('paths', {})
    environment_process = NonSpatialEnvironment(environment)
    processes.update({
        environment_process.name: environment_process})

    # add topology
    environment_topology = environment_process.generate_topology({})[
        environment_process.name]
    environment_topology = deep_merge(
        environment_topology,
        overide_topology)
    topology.update({
        environment_process.name: environment_topology})


def process_in_experiment(
        process: Process,
        settings: Optional[Dict[str, Any]] = None,
        initial_state: Optional[Dict[str, Any]] = None,
) -> Engine:
    """Put a Process in an Engine

    Args:
        process: the Process to put into the Engine
        settings: a dictionary of optional configuration options,
            keywords include timeline, environment, and topology that
            add to or modify the Process.
        initial_state: initial state to overrides the defaults.

    Returns:
        an :term:`Engine`.
    """
    settings = settings or {}
    initial_state = initial_state or {}

    override_topology = settings.get('topology', {})
    processes = {'process': process}
    topology = {
        'process': {
            port: override_topology.get(port, (port,))
            for port in process.ports_schema().keys()}}

    composite = Composite({
        'processes': processes,
        'topology': topology})

    return composite_in_experiment(
        composite=composite,
        settings=settings,
        initial_state=initial_state
    )


def composite_in_experiment(
        composite: Composite,
        settings: Optional[Dict[str, Any]] = None,
        initial_state: Optional[Dict[str, Any]] = None,
) -> Engine:
    """Put a Composite in an Engine

    Args:
        composite: the :term:`Composite` object.
        settings: a dictionary of options, including composite_config
            for configuring the composite. Additional  keywords include
            timeline, environment, and outer_path.
        initial_state: initial state to overrides the defaults.

    Returns:
        an :term:`Engine`.
    """
    settings = settings or {}
    initial_state = initial_state or {}

    processes = composite['processes']
    topology = composite['topology']

    timeline = settings.get('timeline', None)
    if timeline is not None:
        add_timeline(processes, topology, timeline)
        all_times = [t[0] for t in timeline['timeline']]
        settings['total_time'] = max(all_times)

    environment = settings.get('environment', None)
    if environment is not None:
        add_environment(processes, topology, environment)

    # initialize the experiment
    experiment_config = {
        'composite': composite,
        'initial_state': initial_state,
    }
    for key, setting in settings.items():
        if key in experiment_config_keys:
            experiment_config[key] = setting
    return Engine(**experiment_config)  # type: ignore


def composer_in_experiment(
        composer: Composer,
        settings: Optional[Dict[str, Any]] = None,
        initial_state: Optional[Dict[str, Any]] = None,
        config: Optional[Dict[str, Any]] = None,
        outer_path: HierarchyPath = (),
) -> Engine:
    """Generate a Composite in an Engine

    Args:
        composer: a :term:`Composer` object.
        settings: a dictionary of options, including composite_config
            for configuring the composite. Additional  keywords include
            timeline, environment, and outer_path.
        initial_state: initial state to overrides the defaults.
        config: updates values in composer's config
        outer_path: path to the processes and topology

    Returns:
        an :term:`Engine`
    """
    composite = composer.generate(config, outer_path)
    return composite_in_experiment(
        composite=composite,
        settings=settings,
        initial_state=initial_state,
    )


# simulate helper functions

def simulate_process(
        process: Process,
        settings: Optional[Dict[str, Any]] = None
) -> Dict:
    """Put a :term:`Process` in an :term:`Engine` and simulate it"""
    settings = settings or {}
    experiment = process_in_experiment(process, settings)
    return simulate_experiment(experiment, settings)


def simulate_composite(
        composite: Composite,
        settings: Optional[Dict[str, Any]] = None
) -> Dict:
    """Put a :term:`Composite` in an :term:`Engine` and simulate it"""
    settings = settings or {}
    experiment = composite_in_experiment(composite, settings)
    return simulate_experiment(experiment, settings)


def simulate_composer(
        composer: Composer,
        settings: Optional[Dict[str, Any]] = None,
        config: Optional[Dict[str, Any]] = None,
) -> Dict:
    """Initialize a :term:`Composer` in an :term:`Engine` and simulate it"""
    settings = settings or {}
    config = config or {}
    outer_path = settings.get('outer_path', tuple())
    composite = composer.generate(config, outer_path)
    initial_state = settings.get('initial_state')
    experiment = composite_in_experiment(
        composite,
        settings=settings,
        initial_state=initial_state)
    return simulate_experiment(experiment, settings)


def simulate_experiment(
        experiment: Engine,
        settings: Optional[Dict[str, Any]] = None
) -> Dict:
    """Simulate an :term:`Engine`.

    Args:
        experiment: a configured experiment

    Returns:
        A timeseries of variables from all ports. If ``return_raw_data``
        is True, return the raw data instead.
    """
    settings = settings or {}
    total_time = settings.get('total_time', 10)
    return_raw_data = settings.get('return_raw_data', False)

    # run simulation
    experiment.update(total_time)
    experiment.end()

    # return data from emitter
    if return_raw_data:
        return experiment.emitter.get_data()
    return experiment.emitter.get_timeseries()


# Tests
def test_process_in_experiment() -> None:
    process = ExchangeA()
    experiment = process_in_experiment(process)
    assert experiment.processes['process'] is process


def test_process_in_experiment_timeline() -> None:
    timeline = [
        (0, {('internal', 'A'): 0}),
        (1, {('internal', 'A'): 1}),
    ]
    process = ExchangeA()
    experiment = process_in_experiment(
        process,
        settings={
            'timeline': {'timeline': timeline}})
    assert experiment.processes['process'] is process
    assert isinstance(
        experiment.processes['timeline'],
        TimelineProcess)


def test_process_in_experiment_environment() -> None:
    process = ExchangeA()
    experiment = process_in_experiment(
        process,
        settings={'environment': {}})

    assert experiment.processes['process'] is process
    assert isinstance(
        experiment.processes['nonspatial_environment'],
        NonSpatialEnvironment)


def test_composer_in_experiment() -> None:
    composer = PoQo({
        '_schema': {'po': {'A': {'a1': {'_emit': True}}}}})

    timeline = [
        (0, {('aaa', 'a1'): 50}),
        (5, {('aaa', 'a1'): 10}),
        (15, {})]
    settings = {
        'environment': {},
        'timeline': {'timeline': timeline}}
    experiment = composer_in_experiment(composer=composer, settings=settings)

    assert isinstance(
        experiment.processes['nonspatial_environment'],
        NonSpatialEnvironment)

    output = simulate_experiment(experiment, settings)

    # check that timeline worked
    assert output['aaa']['a1'][6] == 10
    assert settings['total_time'] == 15
    assert len(output['aaa']['a1']) == 16


def test_composite_in_experiment() -> None:
    composer = PoQo({
        '_schema': {'po': {'A': {'a2': {'_emit': True}}}}})
    composite = composer.generate()
    settings: Dict = {}
    experiment = composite_in_experiment(
        composite=composite,
        settings=settings)
    assert experiment.processes['po'] is composite['processes']['po']

    output = simulate_composite(composite, settings)
    assert output['aaa']['x'][-1] == -90


def test_process_deletion() -> None:
    """Check that processes are successfully deleted"""
    process = ToyLinearGrowthDeathProcess({'targets': ['process']})
    settings = {
        'emit_step': 1,
        'topology': {
            'global': ('global',),
            'targets': tuple()}}

    output = simulate_process(process, settings)
    expected_masses = [
        # Mass stops increasing the iteration after mass > 5 because
        # cell dies
        1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 7.0, 7.0, 7.0, 7.0]
    masses = output['global']['mass']
    assert masses == expected_masses


def test_composer(return_value: bool = False) -> Optional[Dict]:
    toy_compartment = ToyCompartment({})
    settings = {
        'total_time': 10,
        'initial_state': {
            'periplasm': {
                'GLC': 20,
                'MASS': 100,
                'DENSITY': 10},
            'cytoplasm': {
                'GLC': 0,
                'MASS': 3,
                'DENSITY': 10}}}
    if return_value:
        return simulate_composer(toy_compartment, settings)
    return None


if __name__ == '__main__':
    # test_process_deletion()
    # test_composer()
    # test_replace_names()
    # test_process_in_experiment_environment()
    # test_composer_in_experiment()

    test_composite_in_experiment()  # pragma: no cover



"""
====================
Remove Process
====================
"""

import os
from typing import Any, Dict

from vivarium.core.engine import pp
from vivarium.core.process import (
    Deriver,
)
from vivarium.core.composer import Composer
from vivarium.core.directories import PROCESS_OUT_DIR
from vivarium.core.engine import Engine
from vivarium.composites.toys import ExchangeA
from vivarium.processes.timeline import TimelineProcess

NAME = 'remove'


class Remove(Deriver):
    """ Remove Process

    remove a compartment when the state under the 'trigger' port is set to True.
    """
    name = NAME
    defaults: Dict[str, Any] = {}

    def __init__(self, parameters=None):
        super(Remove, self).__init__(parameters)
        self.agent_id = self.parameters['agent_id']

    def ports_schema(self):
        return {
            'trigger': {
                '_default': False,
                '_emit': True},
            'agents': {'*': {}}}

    def next_update(self, timestep, states):
        if states['trigger']:
            return {
                'agents': {
                    '_delete': [self.agent_id]}}
        else:
            return {}


# test
class ToyLivingCompartment(Composer):
    defaults = {
        'exchange': {'uptake_rate': 0.1},
        'death': {}
    }

    def generate_processes(self, config):
        death_config = config['death']
        death_config['agent_id'] = config['agent_id']
        return {
            'exchange': ExchangeA(config['exchange']),
            'death': Remove(death_config)}

    def generate_topology(self, config):
        agents_path = ('..', '..', 'agents')
        return {
            'exchange': {
                'internal': ('internal',),
                'external': ('external',)},
            'death': {
                # set the trigger to be the 'dead' state
                'trigger': ('dead',),
                'agents': agents_path}}


def test_remove(return_value=False):
    agent_id = '1'

    # timeline turns death on
    time_dead = 5
    time_total = 10
    timeline = [
        (0, {('agents', agent_id, 'dead'): False}),
        (time_dead, {('agents', agent_id, 'dead'): True}),
        (time_total, {})]
    timeline_process = TimelineProcess({'timeline': timeline})

    # make the composite
    compartment = ToyLivingCompartment({
        'agent_id': agent_id})
    composite = compartment.generate(path=('agents', agent_id))
    composite.merge(
        processes={'timeline': timeline_process},
        topology={'timeline': {
            'global': ('global',),
            'agents': ('agents',)}},
    )

    # initial state
    initial_state = {
        'agents': {
            agent_id: {
                'external': {'A': 1},
                'trigger': False}}}

    sim = Engine(
        composite=composite,
        initial_state=initial_state
        )
    sim.update(time_total)
    output = sim.emitter.get_timeseries()

    assert len(output['agents']['1']['dead']) == time_dead + 1
    assert len(output['time']) == time_total + 1

    if return_value:
        return output


def run_remove():
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    os.makedirs(out_dir, exist_ok=True)
    output = test_remove(return_value=True)
    pp(output)


if __name__ == '__main__':
    run_remove()



from vivarium.core.process import Deriver
from vivarium.library.units import units


class CountsToConcentration(Deriver):
    """ Adapts counts to mass-based concentration """
    name = 'counts_to_concentration'
    defaults = {
        'concentration_unit': units.mg / units.mL,
        'default_volume': 1 * units.fL,
        # map molecule name to mw in units.g / units.mol
        'molecular_weights': {},
    }

    def __init__(self, parameters=None):
        parameters = parameters or {}
        if 'molecular_weights' not in parameters:
            parameters['molecular_weights'] = {
                'mass': 1.0 * units.g / units.mol}
        super().__init__(parameters)
        for mol_id, mw in self.parameters['molecular_weights'].items():
            try:
                mw.to(units.g / units.mol)
            except ValueError:
                print(
                    f"{mol_id} needs a mw convertable to units.g / units.mol")

    def initial_state(self, config=None):
        return self.default_state()

    def ports_schema(self):
        keys = list(self.parameters['molecular_weights'].keys())
        return {
            'global': {
                'volume': {
                    '_default': self.parameters['default_volume']}
            },
            'input': {
                key: {
                    '_default': 0,
                } for key in keys
            },
            'output': {
                key: {
                    '_default': 1.0 * self.parameters['concentration_unit'],
                    '_updater': 'set',
                } for key in keys
            }
        }

    def next_update(self, timestep, states):
        counts = states['input']
        volume = states['global']['volume']

        # do conversion
        # Concentration = mass/molecular_weight/characteristic volume
        # Note: here we just set the scale, not the volume
        mass_species_conc = {
            mol_id: (count * units.molec * self.parameters[
                'molecular_weights'][mol_id] / volume).to(
                self.parameters['concentration_unit'])
            for mol_id, count in counts.items()}

        return {'output': mass_species_conc}


class MassToCount(Deriver):
    """ Adapts mass variable to counts """
    name = 'mass_to_count'
    defaults = {
        'input_mass_units': 1.0 * units.fg,
        'molecular_weights': {},
    }

    def __init__(self, parameters=None):
        parameters = parameters or {}
        if 'molecular_weights' not in parameters:
            parameters['molecular_weights'] = {
                'mass': 1.0 * units.fg / units.molec}
        super().__init__(parameters)
        for mol_id, mw in self.parameters['molecular_weights'].items():
            try:
                mw.to(units.g / units.mol)
            except ValueError:
                print(
                    f"{mol_id} needs a mw convertable to units.g / units.mol")

    def initial_state(self, config=None):
        return self.default_state()

    def ports_schema(self):
        keys = list(self.parameters['molecular_weights'].keys())
        return {
            'input': {
                key: {
                    '_default': 1.0 * units.fg,
                } for key in keys
            },
            'output': {
                key: {
                    '_default': 1.0,
                    '_updater': 'set',
                } for key in keys
            },
            'global': {
                'volume': {}
            }
        }

    def next_update(self, timestep, states):
        masses = states['input']

        # do conversion
        # count = mass/molecular_weight
        # Note: here we just set the scale, not the volume
        mass_species_count = {
            mol_id: int((mass / self.parameters[
                'molecular_weights'][mol_id]).magnitude)
            for mol_id, mass in masses.items()}

        return {'output': mass_species_count}



class MassToMolar(Deriver):
    """ Adapts mass variable to mmolar concentration """
    name = 'mass_to_concentration'
    defaults = {
        'mass_units': units.fg,
        'concentration_unit': units.mmolar,
        'default_volume': 1 * units.fL,
        # map molecule name to mw in units.g / units.mol
        'molecular_weights': {},
    }

    def __init__(self, parameters=None):
        parameters = parameters or {}
        if 'molecular_weights' not in parameters:
            parameters['molecular_weights'] = {
                'mass': 1.0 * units.g / units.mol}
        super().__init__(parameters)
        for mol_id, mw in self.parameters['molecular_weights'].items():
            try:
                mw.to(units.g / units.mol)
            except ValueError:
                print(
                    f"{mol_id} needs a mw convertable to units.g / units.mol")

    def initial_state(self, config=None):
        return self.default_state()

    def ports_schema(self):
        keys = list(self.parameters['molecular_weights'].keys())
        return {
            'global': {
                'volume': {
                    '_default': self.parameters['default_volume']}
            },
            'input': {
                key: {
                    '_default': 0 * self.parameters['mass_units'],
                } for key in keys
            },
            'output': {
                key: {
                    '_default': 1.0 * self.parameters['concentration_unit'],
                    '_updater': 'set',
                } for key in keys
            }
        }

    def next_update(self, timestep, states):
        masses = states['input']
        volume = states['global']['volume']

        # do conversion
        # Concentration = mass/molecular_weight/characteristic volume
        # Note: here we just set the scale, not the volume
        mass_species_conc = {
            mol_id: (mass / self.parameters['molecular_weights'][mol_id] /
                     volume).to(self.parameters['concentration_unit'])
            for mol_id, mass in masses.items()}

        return {'output': mass_species_conc}



def test_derivers():
    config = {
        'molecular_weights': {
            'A': 1.0 * units.g / units.mol,
            'B': 2.0 * units.g / units.mol,
        }}

    # MassToCount
    m_to_c = MassToCount(config)

    # convert mass to counts
    mass_in = m_to_c.initial_state()
    mass_in['input'] = {'A': 10 * units.fg, 'B': 10 * units.fg}
    counts_out = m_to_c.next_update(0, mass_in)

    # MassToConcentration
    m_to_conc = CountsToConcentration(config)

    # convert mass to concentration
    mass_in = m_to_conc.initial_state()
    mass_in['input'] = {'A': 10, 'B': 10}
    concs_out = m_to_conc.next_update(0, mass_in)

    # asserts
    assert counts_out == {'output': {'A': 10, 'B': 5}}
    # TODO assert concs_out
    _ = concs_out

if __name__ == '__main__':
    test_derivers()



"""
==========
Serialize
==========

Collection of serializers that transform Python data into
a BSON-compatible form.
"""

import re
import math
import warnings
from typing import Any, List, Optional, Union
from collections.abc import Callable

import orjson
import numpy as np
from pint import Unit  # pylint: disable=unused-import
try:
    from pint.quantity import Quantity  # pylint: disable=unused-import
except ImportError:
    from pint import Quantity  # pylint: disable=unused-import
from vivarium.core.process import Process
from vivarium.library.units import units
from vivarium.core.registry import serializer_registry, Serializer


def find_numpy_and_non_strings(
    d: dict,
    curr_path: tuple=tuple(),
    saved_paths: Optional[List[tuple]]=None
) -> List[tuple]:
    """Return list of paths which terminate in a non-string or Numpy string
    dictionary key. Orjson does not handle these types of keys by default."""
    if not saved_paths:
        saved_paths = []
    if isinstance(d, dict):
        for key in d.keys():
            if not isinstance(key, str):
                saved_paths.append(curr_path + (key,))
            elif isinstance(key, np.str_):
                saved_paths.append(curr_path + (key,))
            saved_paths = find_numpy_and_non_strings(
                d[key], curr_path+(key,), saved_paths)
    return saved_paths


def serialize_value(
    value: Any,
    default: Optional[Callable]=None,
) -> Any:
    """Apply orjson-based serialization routine on ``value``.

    Args:
        value (Any): Data to be serialized. All keys must be strings. Notably,
            Numpy strings (``np.str_``) are not acceptable keys.
        default (Callable): A function that is called on any data of a type
            that is not natively supported by orjson. Returns an object that
            can be handled by default up to 254 times before an exception is
            raised.

    Returns:
        Any: Serialized data
    """
    if default is None:
        default = make_fallback_serializer_function()
    try:
        value = orjson.dumps(value, option=orjson.OPT_SERIALIZE_NUMPY,
            default=default)
        return orjson.loads(value)
    except TypeError as e:
        bad_keys = find_numpy_and_non_strings(value)
        raise TypeError('These paths end in incompatible non-string or Numpy '
            f'string keys: {bad_keys}').with_traceback(e.__traceback__) from e


def deserialize_value(value: Any) -> Any:
    """Find and apply the correct serializer for a value
    by calling each registered serializer's
    :py:meth:`vivarium.core.registry.Serializer.can_deserialize()`
    method. Returns the value as is if no compatible serializer
    is found.

    Args:
        value (Any): Data to be deserialized

    Raises:
        ValueError: Only one serializer should apply for any given value

    Returns:
        Any: Deserialized data
    """
    compatible_serializers = []
    for serializer_name in serializer_registry.list():
        serializer = serializer_registry.access(serializer_name)
        if serializer.can_deserialize(value):
            compatible_serializers.append(serializer)
    if not compatible_serializers:
        # Most likely a built-in type with no custom serializer/deserializer
        return value
    if len(compatible_serializers) > 1:
        raise ValueError(
            f'Multiple deserializers ({compatible_serializers}) found '
            f'for {value}')
    serializer = compatible_serializers[0]
    return serializer.deserialize(value)


class SequenceDeserializer(Serializer):
    """Iterates through lists and applies deserializers.
    """
    python_type = list

    def can_deserialize(self, data: Any) -> bool:
        return isinstance(data, list)

    def deserialize(self, data: Any) -> List[Any]:
        return [deserialize_value(value) for value in data]


class DictDeserializer(Serializer):
    """Iterates through dictionaries and applies deserializers.
    """
    python_type = dict

    def can_deserialize(self, data: Any) -> bool:
        return isinstance(data, dict)

    def deserialize(self, data: dict) -> dict:
        return {
            key: deserialize_value(value)
            for key, value in data.items()
        }


class NumpyFallbackSerializer(Serializer):
    """Orjson does not handle Numpy arrays with strings
    """
    python_type = np.ndarray

    def serialize(self, data: Any) -> list:
        return data.tolist()


class UnitsSerializer(Serializer):
    """Serializes data with units into strings of the form ``!units[...]``,
    where ``...`` is the result of calling ``str(data)``. Deserializes strings
    of this form back into data with units."""

    def __init__(self) -> None:
        super().__init__()
        self.regex_for_serialized = re.compile('!units\\[(.*)\\]')

    python_type = type(units.fg)

    def serialize(self, data: Any) -> Union[List[str], str]:
        try:
            return_value = []
            for subvalue in data:
                return_value.append(f"!units[{str(subvalue)}]")
            return return_value
        except TypeError:
            return f"!units[{str(data)}]"

    def can_deserialize(self, data: Any) -> bool:
        if not isinstance(data, str):
            return False
        return bool(self.regex_for_serialized.fullmatch(data))

    # Here the differing argument is `unit`, which is optional, so we
    # can ignore the pylint warning.
    def deserialize(self, data: str, unit=None):  # type: ignore  # pylint: disable=arguments-differ
        """Deserialize data with units from a human-readable string.

        Args:
            data: The data to deserialize. Providing a list here is
                deprecated. You should use ``deserialize_value``
                instead, which uses a separate list deserializer.
            unit: The units to convert ``data`` to after deserializing.
                If omitted, no conversion occurs. This option is
                deprecated.

        Returns:
            A single deserialized object or, if ``data`` is a list, a
            list of deserialized objects.
        """
        if unit is not None:
            warnings.warn(
                'The `unit` argument to `UnitsSerializer.deserialize` is '
                'deprecated.',
                DeprecationWarning,
            )
        if isinstance(data, list):
            warnings.warn(
                'Passing a list to `UnitsSerializer.deserialize` is '
                'deprecated. Please use `deserialize_value()` instead.',
                DeprecationWarning,
            )
            unit_data = [units(d) for d in data]
            if unit is not None:
                unit_data = [d.to(unit) for d in data]
        else:
            # Extract ... from !units[...].
            matched_regex = self.regex_for_serialized.fullmatch(data)
            if matched_regex:
                data = matched_regex.group(1)
            if data.startswith('nan'):
                unit_str = data[len('nan'):].strip()
                unit_data = math.nan * units(unit_str)
            else:
                unit_data = units(data)
            if unit is not None:
                unit_data.to(unit)  # type: ignore
        return unit_data


class QuantitySerializer(Serializer):
    """Serializes data with units into strings of the form ``!units[...]``,
    where ``...`` is the result of calling ``str(data)``. Deserializes strings
    of this form back into data with units."""
    python_type = type(1*units.fg)

    def serialize(self, data: Any) -> Union[List[str], str]:
        try:
            return_value = []
            for subvalue in data:
                return_value.append(f"!units[{str(subvalue)}]")
            return return_value
        except TypeError:
            return f"!units[{str(data)}]"


class SetSerializer(Serializer):
    """Serializer for set objects."""
    python_type = set

    def serialize(self, data: set) -> List:
        return list(data)


class FunctionSerializer(Serializer):
    """Serializer for function objects."""
    python_type = type(deserialize_value)

    def serialize(self, data: Callable) -> str:
        return f"!FunctionSerializer[{str(data)}]"


class ProcessSerializer(Serializer):
    """Serializer for processes if ``emit_process`` is enabled."""
    python_type = Process

    def serialize(self, data: Process) -> str:
        proc_str = str(dict(data.parameters, _name=data.name))
        return f"!ProcessSerializer[{proc_str}]"


def make_fallback_serializer_function() -> Callable:
    """Creates a fallback function that is called by orjson on data of
    types that are not natively supported. Define and register instances of
    :py:class:`vivarium.core.registry.Serializer()` with serialization
    routines for the types in question."""

    def default(obj: Any) -> Any:
        # Try to lookup by exclusive type
        serializer = serializer_registry.access(str(type(obj)))
        if not serializer:
            compatible_serializers = []
            for serializer_name in serializer_registry.list():
                test_serializer = serializer_registry.access(serializer_name)
                # Subclasses with registered serializers will be caught here
                if isinstance(obj, test_serializer.python_type):
                    compatible_serializers.append(test_serializer)
            if len(compatible_serializers) > 1:
                raise TypeError(
                    f'Multiple serializers ({compatible_serializers}) found '
                    f'for {obj} of type {type(obj)}')
            if not compatible_serializers:
                raise TypeError(
                    f'No serializer found for {obj} of type {type(obj)}')
            serializer = compatible_serializers[0]
            if not isinstance(obj, Process):
                # We don't warn for processes because since their types
                # based on their subclasses, it's not possible to avoid
                # searching through the serializers.
                warnings.warn(
                    f'Searched through serializers to find {serializer} '
                    f'for data of type {type(obj)}. This is '
                    f'inefficient.')
        return serializer.serialize(obj)
    return default



"""
===================================
Toy Glucose Phosphorylation Process
===================================

This is a toy example referenced in the documentation.
"""

from vivarium.core.process import Process
from vivarium.core.engine import Engine
from vivarium.plots.simulation_output import plot_simulation_output
from vivarium.library.units import units


class GlucosePhosphorylation(Process):

    name = 'glucose_phosphorylation'
    defaults = {
        'k_cat': 2e-3,
        'K_ATP': 5e-2,
        'K_GLC': 4e-2,
    }

    def ports_schema(self):
        return {
            'cytoplasm': {
                'GLC': {
                    # accumulate means to add the updates
                    '_updater': 'accumulate',
                    '_default': 1.0,
                    '_properties': {
                        'mw': 1.0 * units.g / units.mol,
                    },
                    '_emit': True,
                },
                # accumulate is the default, so we don't need to specify
                # updaters for the rest of the variables
                'G6P': {
                    '_default': 0.0,
                    '_properties': {
                        'mw': 1.0 * units.g / units.mol,
                    },
                    '_emit': True,
                },
                'HK': {
                    '_default': 0.1,
                    '_properties': {
                        'mw': 1.0 * units.g / units.mol,
                    },
                },
            },
            'nucleoside_phosphates': {
                'ATP': {
                    '_default': 2.0,
                    '_emit': True,
                },
                'ADP': {
                    '_default': 0.0,
                    '_emit': True,
                }
            },
        }

    def next_update(self, timestep, states):
        # Get concentrations from state
        cytoplasm = states['cytoplasm']
        nucleoside_phosphates = states['nucleoside_phosphates']
        hk = cytoplasm['HK']
        glc = cytoplasm['GLC']
        atp = nucleoside_phosphates['ATP']

        # Get kinetic parameters
        k_cat = self.parameters['k_cat']
        k_atp = self.parameters['K_ATP']
        k_glc = self.parameters['K_GLC']

        # Compute reaction rate with michaelis-menten equation
        rate = k_cat * hk * glc * atp / (
            k_glc * k_atp + k_glc * atp + k_atp * glc + glc * atp)

        # Compute concentration changes from rate and timestep
        delta_glc = -rate * timestep
        delta_atp = -rate * timestep
        delta_g6p = rate * timestep
        delta_adp = rate * timestep

        # Compile changes into an update
        update = {
            'cytoplasm': {
                'GLC': delta_glc,
                'G6P': delta_g6p,
                # We exclude HK because it doesn't change
            },
            'nucleoside_phosphates': {
                'ATP': delta_atp,
                'ADP': delta_adp,
            },
        }

        return update


if __name__ == '__main__':
    total_time = 10
    parameters = {
        'k_cat': 1.5,
        'time_step': 0.1,
    }
    my_process = GlucosePhosphorylation(parameters)
    my_composite = my_process.generate()
    sim = Engine(
        processes=my_composite['processes'],
        topology=my_composite['topology']
    )
    sim.update(total_time)
    timeseries = sim.emitter.get_timeseries()
    plot_simulation_output(timeseries, {}, './')



"""
==============================================
Composite, Composer, and MetaComposer Classes
==============================================
"""

import abc
import copy
from typing import Dict, Any, Optional, Iterable, List

from vivarium.core.process import (
    _override_schemas, assoc_in, _get_parameters, Process)
from vivarium.core.store import (
    Store, generate_state)
from vivarium.core.types import (
    Processes, Topology, HierarchyPath, State, Schema, Steps, Flow)
from vivarium.library.datum import Datum
from vivarium.library.dict_utils import (
    deep_merge, deep_merge_check, deep_copy_internal)
from vivarium.library.topology import inverse_topology


def _get_composite_state_recur(
        processes: Processes,
        steps: Steps,
        topology: Any,
        state_type: Optional[str] = 'initial',
        path: Optional[HierarchyPath] = None,
        config: Optional[dict] = None,
) -> Optional[State]:
    path = path or tuple()
    config = config or {}
    processes = processes or {}
    steps = steps or {}
    topology = topology or {}

    state: dict = {}
    all_keys = set(processes.keys() | steps.keys())
    for key in all_keys:
        sub_path: HierarchyPath = path + (key,)
        sub_topology: Any = topology.get(key)
        sub_processes: Any = processes.get(key)
        sub_steps: Any = steps.get(key)
        sub_state: Any = None

        if isinstance(sub_processes, dict) or isinstance(sub_steps, dict):
            sub_state = _get_composite_state_recur(
                processes=sub_processes,
                steps=sub_steps,
                topology=sub_topology,
                state_type=state_type,
                path=sub_path,
                config=config.get(key),
            )
        elif isinstance(sub_processes, Process) or \
                isinstance(sub_steps, Process):
            node: Process = sub_processes or sub_steps
            if state_type == 'initial':
                process_state = node.initial_state(config.get(node.name))
            elif state_type == 'default':
                process_state = node.default_state()
            # Prevent multiupdates from forming when a single process has
            # multiple ports to the same stores holding a dictionary
            sub_state = inverse_topology(
                path, process_state, sub_topology, multi_updates=False)
        else:
            raise ValueError(f'invalid processes {sub_processes} or steps'
                             + str(sub_steps))
        state = deep_merge(state, sub_state)
    return state


def _get_composite_state(
        processes: Processes,
        steps: Steps,
        topology: Any,
        state_type: Optional[str] = 'initial',
        path: Optional[HierarchyPath] = None,
        initial_state: Optional[State] = None,
        config: Optional[dict] = None,
) -> Optional[State]:
    state = _get_composite_state_recur(
        processes=processes,
        steps=steps,
        topology=topology,
        state_type=state_type,
        path=path,
        config=config,
    )
    state = deep_merge(state, initial_state)
    return state


class Composite(Datum):
    """Composite parent class.

    Contains keys for processes and topology
    """
    processes: Processes = {}
    steps: Steps = {}
    flow: Flow = {}
    topology: Topology = {}
    state: State = {}
    defaults: Dict[str, Any] = {
        'processes': {},
        'steps': {},
        'flow': {},
        'topology': {},
        'state': {},
    }

    def __init__(
            self,
            config: Optional[Dict[str, Any]] = None,
            store: Optional[Store] = None,
            processes: Optional[Processes] = None,
            steps: Optional[Steps] = None,
            flow: Optional[Flow] = None,
            topology: Optional[Topology] = None,
            state: Optional[State] = None,
    ) -> None:
        if not config:
            if store:
                composite = get_composite_from_store(store)
                processes = composite.processes
                topology = composite.topology
                steps = composite.steps
                flow = composite.flow
                state = composite.state
            config = {
                'processes': processes or {},
                'topology': topology or {},
                'steps': steps or {},
                'flow': flow or {},
                'state': state or {}
            }
        super().__init__(config)
        self._schema = config.get('_schema', {})
        processes_and_steps = deep_copy_internal(self.processes)
        deep_merge_check(processes_and_steps, self.steps)
        _override_schemas(self._schema, processes_and_steps)

    def generate_store(self, config: Optional[dict] = None) -> Store:
        config = config or {}
        initial_state = self.initial_state(config)
        return generate_state(
            processes=self.processes,
            steps=self.steps,
            topology=self.topology,
            flow=self.flow,
            initial_state=initial_state)

    def initial_state(self, config: Optional[dict] = None) -> Optional[State]:
        """ Merge all processes' initial states
        Arguments:
            config (dict): A dictionary of configuration options. All
            subclass implementation must accept this parameter, but
            some may ignore it.
        Returns:
            (dict): Subclass implementations must return a dictionary
            mapping state paths to initial values.
        """
        config = config or {}
        initial_state = config.get('initial_state', {})
        initial_state = deep_merge(copy.deepcopy(self.state), initial_state)
        return _get_composite_state(
            processes=self.processes,
            steps=self.steps,
            topology=self.topology,
            state_type='initial',
            initial_state=initial_state,
            config=config)

    def default_state(self, config: Optional[dict] = None) -> Optional[State]:
        """ Merge all processes' default states
        Arguments:
            config (dict): A dictionary of configuration options. All
            subclass implementation must accept this parameter, but
            some may ignore it.
        Returns:
            (dict): Subclass implementations must return a dictionary
            mapping state paths to default values.
        """
        return _get_composite_state(
            processes=self.processes,
            steps=self.steps,
            topology=self.topology,
            state_type='default',
            config=config)

    def merge(
            self,
            composite: Optional['Composite'] = None,
            processes: Optional[Dict[str, 'Process']] = None,
            topology: Optional[Topology] = None,
            steps: Optional[Steps] = None,
            flow: Optional[Flow] = None,
            state: Optional[State] = None,
            path: Optional[HierarchyPath] = None,
            schema_override: Optional[Schema] = None,
    ) -> None:
        composite = composite or Composite({})
        processes = processes or {}
        topology = topology or {}
        steps = steps or {}
        flow = flow or {}
        path = path or tuple()
        state = state or {}
        schema_override = schema_override or {}

        # get the processes and topology to merge
        merge_processes = {}
        merge_topology = {}
        merge_steps = {}
        merge_flow = {}
        merge_state = {}
        if composite:
            merge_processes.update(composite['processes'])
            merge_topology.update(composite['topology'])
            merge_steps.update(composite['steps'])
            merge_flow.update(composite['flow'])
            merge_state.update(composite.get('state', {}))

        deep_merge(merge_processes, processes)
        deep_merge(merge_topology, topology)
        deep_merge(merge_steps, steps)
        deep_merge(merge_flow, flow)
        deep_merge(merge_state, state)
        merge_processes = assoc_in({}, path, merge_processes)
        merge_topology = assoc_in({}, path, merge_topology)
        merge_steps = assoc_in({}, path, merge_steps)
        merge_flow = assoc_in({}, path, merge_flow)
        merge_state = assoc_in({}, path, merge_state)

        # merge with instance processes and topology
        deep_merge(self.processes, merge_processes)
        deep_merge(self.topology, merge_topology)
        deep_merge(self.steps, merge_steps)
        deep_merge(self.flow, merge_flow)
        deep_merge(self.state, merge_state)
        self._schema.update(schema_override)

        processes_and_steps = deep_copy_internal(self.processes)
        deep_merge_check(processes_and_steps, self.steps)
        _override_schemas(self._schema, processes_and_steps)

    def get_parameters(self) -> Dict:
        """Get the parameters for all :term:`processes`.
        Returns:
            A map from process names to parameters.
        """
        return _get_parameters(self.processes)


def get_composite_from_store(store: Store) -> Composite:
    """Make a :term:`Composite` from a :term:`Store`"""
    return Composite(
        processes=store.get_processes(),
        topology=store.get_topology(),
        steps=store.get_steps(),
        flow=store.get_flow(),
        state=store.get_value(),
    )


class Composer(metaclass=abc.ABCMeta):
    defaults: Dict[str, Any] = {}

    def __init__(self, config: Optional[dict] = None) -> None:
        """Base class for :term:`composer` classes.

        Composers generate :term:`composites`.

        All :term:`composer` classes must inherit from this class.

        Args:
            config: Dictionary of configuration options that can
                override the class defaults.
        """
        config = config or {}
        if 'name' in config:
            self.name = config['name']
        elif not hasattr(self, 'name'):
            self.name = self.__class__.__name__

        self.config = copy.deepcopy(self.defaults)
        self.config = deep_merge(self.config, config)
        self.schema_override = self.config.pop('_schema', {})

    def generate_store(self, config: Optional[dict] = None) -> Store:
        composite = self.generate()
        return composite.generate_store(config)

    @abc.abstractmethod
    def generate_processes(
            self,
            config: Optional[dict]) -> Processes:
        """Generate processes dictionary.

        Every subclass must override this method. For backwards
        compatibility, :py:class:`vivarium.core.process.Step` objects
        may be included in the returned dictionary, but this practice is
        discouraged and may be disallowed in a future release.

        Args:
            config: A dictionary of configuration options. All
                subclass implementation must accept this parameter, but
                some may ignore it.

        Returns:
            Subclass implementations must return a dictionary
            mapping process names to instantiated and configured
            :py:class:`vivarium.core.process.Process` objects.
        """
        return {}  # pragma: no cover

    def generate_steps(self, config: Optional[dict]) -> Steps:
        '''Generate the steps dictionary.

        Subclasses that want to include :term:`steps` should override
        this method. This method is the preferred way to specify steps,
        though they may also be returned by
        :py:meth:`generate_processes`.

        Args:
            config: A dictionary of configuration options. All
                subclass implementation must accept this parameter, but
                some may ignore it.

        Returns:
            Subclass implementations should return a dictionary mapping
            step names to instantiated and configured
            :py:class:`vivarium.core.process.Step` objects.
        '''
        _ = config
        _ = self # Silence pylint no-self-use
        return {}  # pragma: no cover

    def generate_flow(self, config: Optional[dict]) -> Flow:
        '''Generate the flow of :term:`step` dependencies.

        Args:
            config: A dictionary of configuration options. All
                subclass implementation must accept this parameter, but
                some may ignore it.

        Returns:
            Subclass implementations should return a dictionary mapping
            step names to sequences (e.g. lists or tuples) of
            :term:`paths`. **Steps with no dependencies must be
            included,** but they should be mapped to an empty sequence.
            Any steps returned by :py:meth:`generate_steps` or
            :py:meth:`generate_processes` that are not included in the
            flow will be treated as if they depend on every step
            previously added to the :term:`engine`.
        '''
        _ = config
        _ = self # Silence pylint no-self-use
        return {}  # pragma: no cover

    @abc.abstractmethod
    def generate_topology(self, config: Optional[dict]) -> Topology:
        """Generate topology dictionary.

        Every subclass must override this method.

        Args:
            config: A dictionary of configuration options. All
                subclass implementation must accept this parameter, but
                some may ignore it.

        Returns:
            Subclass implementations must return a :term:`topology`
            dictionary.
        """
        return {}  # pragma: no cover

    def generate(
            self,
            config: Optional[dict] = None,
            path: HierarchyPath = ()) -> Composite:
        """Generate processes and topology dictionaries.

        Args:
            config: Updates values in the configuration declared
                in the constructor.
            path: Tuple with ('path', 'to', 'level') associates
                the processes and topology at this level.

        Returns:
            Dictionary with the following keys

            * ``processes``: Generated by :py:meth:`generate_processes`.
            * ``steps``: Generated by :py:meth:`generate_steps`.
            * ``flow``: Generated by :py:meth:`generate_flow`.
            * ``topology``: Generated by :py:meth:`generate_topology`.

            The values of these keys are all dictionaries suitable to be
            passed to the constructor for
            :py:class:`vivarium.core.engine.Engine`.
        """
        if config is None:
            config = self.config
        else:
            default = copy.deepcopy(self.config)
            config = deep_merge(default, config)

        processes = self.generate_processes(config)
        steps = self.generate_steps(config)
        flow = self.generate_flow(config)
        topology = self.generate_topology(config)
        processes_and_steps = deep_copy_internal(processes)
        deep_merge_check(processes_and_steps, steps)
        _override_schemas(self.schema_override, processes_and_steps)

        return Composite({
            'processes': assoc_in({}, path, processes),
            'steps': assoc_in({}, path, steps),
            'flow': assoc_in({}, path, flow),
            'topology': assoc_in({}, path, topology),
        })

    def initial_state(self, config: Optional[dict] = None) -> Optional[State]:
        """ Merge all processes' initial states

        Every subclass may override this method.

        Arguments:
            config (dict): A dictionary of configuration options. All
                subclass implementation must accept this parameter, but
                some may ignore it.

        Returns:
            dict: Subclass implementations must return a dictionary
            mapping state paths to initial values.
        """
        composite = self.generate(config)
        return composite.initial_state(config)

    def get_parameters(self) -> dict:
        """Get the parameters for all :term:`processes`.

        Returns:
            A map from process names to dictionaries of those processes'
            parameters.
        """
        composite = self.generate()
        return composite.get_parameters()


class MetaComposer(Composer):

    def __init__(
            self,
            composers: Iterable[Any] = tuple(),
            config: Optional[dict] = None,
    ) -> None:
        """A collection of :py:class:`Composer` objects.

        The MetaComposer can be used to create composites that combine
        all the composers in the collection.

        Args:
            composers: Initial collection of composers.
            config: Initial configuration.
        """
        super().__init__(config)
        self.composers: List = list(composers)

    def _generate(
            self, config: Optional[dict], method: str
    ) -> Dict[str, Any]:
        combined: Dict = {}
        for composer in self.composers:
            func = getattr(composer, method)
            composer_config = copy.deepcopy(composer.config)
            deep_merge(composer_config, config)
            new = func(composer_config)
            if set(combined.keys()) & set(new.keys()):
                raise ValueError(
                    f"{set(combined.keys())} and "
                    f"{set(new.keys())} "
                    f"contain overlapping keys. "
                    f"They were produced by Composer.{method}()")
            combined.update(new)
        return combined

    def generate_processes(
            self,
            config: Optional[dict] = None
    ) -> Dict[str, Any]:
        return self._generate(config, 'generate_processes')

    def generate_steps(
            self,
            config: Optional[dict] = None
    ) -> Dict[str, Any]:
        return self._generate(config, 'generate_steps')

    def generate_flow(
            self,
            config: Optional[dict] = None
    ) -> Dict[str, Any]:
        return self._generate(config, 'generate_flow')

    def generate_topology(
            self,
            config: Optional[dict] = None
    ) -> Topology:
        return self._generate(config, 'generate_topology')

    def add_composer(
            self,
            composer: Composer,
            config: Optional[Dict] = None,
    ) -> None:
        """Add a composer to the collection of stored composers.

        Args:
             composer: The composer to add.
             config: The composer's configuration, which will be merged
                 with the stored config.
        """
        if config:
            self.config.update(config)
        self.composers.append(composer)

    def add_composers(
            self,
            composers: List,
            config: Optional[Dict] = None,
    ) -> None:
        """Add multiple composers to the collection of stored composers.

        Args:
            composers: The composers to add.
            config: Configuration for the composers, which will be
                merged with the stored config.
        """
        if config:
            self.config.update(config)
        self.composers.extend(composers)
